--[[--------------------------------------------------------------------------
  Developed by Nathan Pieper - nrpieper (@) gmail (dot) com
  Misspelled is an interactive chat text spell checker for World of Warcraft

  This code freely distributed for your use in any GPL compliant project.
  Portions of this code are covered by the Gnu Plublic License (GPL)
  Dictionaries use the OpenOffice/HunSpell/ASpell dictionary format
  (http://wiki.services.openoffice.org/wiki/Dictionaries)
  Phonetic cache information created via the NetSpell dictionary util.
  References & Credits:
  NetSpell: http://www.loresoft.com/projects/netspell
  Metaphone Algorithm http://aspell.net/metaphone/
  Edit Distance Algorithm http://en.wikipedia.org/wiki/Levenshtein_distance
  Portions Copyright (c) 2003, Paul Welter All rights reserved.

--------------------------------------------------------------------------]]--


--[[
Challenges:


Issue #1
(http://www.wowwiki.com/ItemLink)
Color tags can be distroyed or mangled as you perform edit near the start
and end of the color tag.

If you press delete with the cursor positioned, just before, a color tag,
the start of the tag (|cffffffff) and the next character are removed.  The
closeing color tag (|r) is left at the end.

If you press backspace with the cursor positioned just past the (|r) end
of a color tag, the (|r) is removed along with the character just before it.

If you press delete with the cursor positioned, just before the last char,
in a color tagged string, the character and the closing (|r) are both removed.

If you press backspace with the cursor positioned, just after the first char,
in a color tagged string, the first char and the beginning color tag are both removed.

In the case of colored item links, Wow just deletes the entire item, if you
hit a delete or backspace.

Solution:
First parse the line to flag any valid item links or textures so they don't get distroyed by the next step.
Then if there are any remaining start or end color tag, remove them.

Issue #2
The Chat Edit box's edit cursor position is relative to characters that don't display in the
chat edit text, i.e. item links, or color tags.
When we add a color tag, or remove one and replace a given misspelled word with a replacement
suggestion of a different length, we should adjust the cursor position so it acts natural.
An OnCursorChanged event doesn't exist or doesn't fire for the ChatEditBox.

Solution:
Use the ChatFrameEditBox's OnUpdate event to detect if the cursor is in a new location.
Store this information, in the local: CurrentCursorPosition, updating it only when the cursor changes
positions.
Methods that insert or delete text from the ChatEditBox can use this saved value to properly adjust
the cursor position as needed when, visible or hidden, characters are insertted or deletted to
the left of the cursor.

--]]--


Misspelled = {
	VERSION = "Misspelled 0.1.15"
}
local _G = _G
local SPELLED_WRONG_HIGHLIGHT = "|cffff91c8" --The color misspelled words will get changed into.
local WordCache = {}  --Stores a cache of every word checked, along with the suggestions for words misspelled
local WordCacheCount = 0  --Counter used to track when we should clean the WordCache table to save memory
local WordCacheCountMax = 5000 --Number of entries tha can live in the WordCache before we clean the cache
local WordLocations = {} --Lookup table used by multiple functions to determin where each work starts and ends
local RightClickedWord = nil --The current word under the CursorPosition that was right-clicked
local RightClickedWordStartPos --Where that word starts
local RightClickedWordStartPos --and where that word ends
local PositionTrack = nil
local OldCursorPosition
local CurrentCursorPosition
local OldLineLength  --Tracks the previous length of the ChatEditBox.text
local GuildRosterCalled = false

function Misspelled_OnLoad()
	this:RegisterEvent("VARIABLES_LOADED")
	this:RegisterEvent("GUILD_ROSTER_UPDATE")
end

function Misspelled_OnEvent()
	if ( event == "VARIABLES_LOADED" ) then
		Misspelled_VariablesLoaded()
	elseif( event == "GUILD_ROSTER_UPDATE") then
		Misspelled:LoadGuildAndFriendRoster()
	end
end

function Misspelled_VariablesLoaded()
	WordDict:Init()
	print("Misspelled: Dict Loaded")

	--Load user dict
	Misspelled:LoadUserDict()

	--Guild members and Friends are valid words.
	--Wait for the GUILD_ROSTER_UPDATE event to load
	--the guild members & friends into the dictionary as valid words.
	--GuildRoster can only be called every so often.
	--If another addon triggered it first, we might miss notification.

	--Moved to OnTextChanged
	--GuildRoster()


	hooksecurefunc("ChatEdit_OnEscapePressed", Misspelled_OnEscapePressed)
	hooksecurefunc("ChatEdit_OnEnterPressed", Misspelled_OnEnterPressed)
	hooksecurefunc("ChatEdit_OnTextChanged", Misspelled_OnTextChanged)
	hooksecurefunc("ChatEdit_OnUpdate", Misspelled_OnUpdate)
	ChatFrameEditBox:SetScript("OnMouseUp", Misspelled_OnMouseUp)

	--No luck here.  ChatFrameEditBox doesn't appear to fire OnKeyDown
	--ChatFrameEditBox:SetScript("OnKeyDown", Misspelled_OnKeyDown)

	-- hooks for removing any misspelled word highlighting in the text before the chat message is sent
	-- The Wow client will disconnect if you attempt to send a colored chat message.
	Misspelled_Orig_SendChatMessage = SendChatMessage;
	SendChatMessage = Misspelled_SendChatMessage;
end



--Before a chat message is sent, remove any highlighting that Misspelled might have added.
--The Wow client will disconnect if you attempt to send a colored chat message.
function Misspelled_SendChatMessage(message, type, language, channel)
	local newText = Misspelled:RemoveHighlighting(message)

	WordLocations = {}
	PositionTrack = nil

	Misspelled_Orig_SendChatMessage(newText, type, language, channel);
end


--Detect when the cursor moves positions
function Misspelled_OnUpdate()
	--print("ChatEditBox Visible", ChatFrameEditBox:GetCursorPosition())
	CurrentCursorPosition = ChatFrameEditBox:GetCursorPosition()
	if CurrentCursorPosition ~= OldCursorPosition then
		--print("New Cursor Position:", CurrentCursorPosition)
		Misspelled_OnCursorChanged(OldCursorPosition, CurrentCursorPosition)
		OldCursorPosition = CurrentCursorPosition
	end
end

function Misspelled_OnCursorChanged(oldPosition, currentPosition)
	---
end


--Possible changes we need to deal with.
--A single new character was inserted at the end of the line (line length grew by 1)
--	Check if this char is a word seperator, then spell check the line
--
--Any other change in the line, requires we recheck the entire line.
--More than a single character was pasted or linked into the line.
--	remove any highlighting and recheck the entire line.
--
--A single character was removed from the line
function Misspelled_OnTextChanged()
	--print ("TextChaged:", ChatFrameEditBox:GetCursorPosition(), ChatFrameEditBox:GetText())

	--track typing in the ChatEdit box, looking for word boundries.
	--spell check new words, tracking suggestions for misspelled words

	--Load the guild roster if needed
	if GuildRosterCalled == false then
		if IsInGuild() == 1 then
			GuildRoster()
		else
			Misspelled:LoadGuildAndFriendRoster()
		end
		GuildRosterCalled = true
	end

	local text = ChatFrameEditBox:GetText()
	local pos = ChatFrameEditBox:GetCursorPosition()

	local newLineLength = #text


	--if the first char is a /, incidating some command, do no spellchecking.
	if (strsub(text, 1, 1) == "/" ) then
		local newText = Misspelled:RemoveHighlighting(text)
		if text ~= newText then
			ChatFrameEditBox:SetText(newText)
			if pos == 1 or pos== 0  then
			  ChatFrameEditBox:SetCursorPosition(pos)
			end
			OldLineLength = #text
		end

		RightClickedWord = nil
		WordLocations = {}
		return
	end

	--If we currently just added one new char to the end of the line, see if it's a word boundry char
	--If so, spell check the line
	if pos == #text and newLineLength -1 == OldLineLength then
		local lastChar = text:sub(-1, -1)
		if string.match(lastChar, "[ ,\.!\?:\"]") ~= nil then
			Misspelled:SpellCheckChat()
		end
	else
		----We must be making some other kind of edit.
		----Recheck the entire line
		Misspelled:SpellCheckChat()

		--ChatFrameEditBox:SetCursorPosition(0)
		--ChatFrameEditBox:SetCursorPosition(pos)
	end

--~ 	--If there was a change in the size of the line, adjust the
--~ 	--cursor position by the # of characters that have been added or removed.
--~ 	if newLineLength ~= #ChatFrameEditBox:GetText() then
--~ 		if OldLineLength ~= nil then
--~ 			ChatFrameEditBox:SetCursorPosition(pos - OldLineLength - #ChatFrameEditBox:GetText() + 1)
--~ 		end
--~ 	end

	--Save the current line length for use with the next round of OnTextChanged processing.
	OldLineLength = #ChatFrameEditBox:GetText()
end

function Misspelled_OnEscapePressed()
	RightClickedWord = nil
	WordLocations = {}
end

function Misspelled_OnEnterPressed()
	RightClickedWord = nil
	WordLocations = {}

	--before message is sent, remove and misspelled highlighting
	local newText = Misspelled:RemoveHighlighting(ChatFrameEditBox:GetText())
	ChatFrameEditBox:SetText(newText)
end


--Return a string where misspelled words are highlighted
--If this string is a slash command "/", don't perform any highlighting
--We should keep track of the current edit cursor position,
--so we can report it's new position after highlighting.
function Misspelled:SpellCheckChat()
	local text = ChatFrameEditBox:GetText()

	if #text < 2 then return end

	--Ensure we have hooked the MouseUp event
	--Should fix Chatter changing this to nil
	if ChatFrameEditBox:GetScript("OnMouseUp") ~= Misspelled_OnMouseUp then
		ChatFrameEditBox:SetScript("OnMouseUp", Misspelled_OnMouseUp)
	end

	local newText = text

	--Watch how many characters we insert or remove from the left side of the cursor position.
	--We'll adjust it's position later to compensate for any "hidden" tags we added.
	local cPos = CurrentCursorPosition
	local charsAdded = 0
	local newCPos

	--remove any previous misspelling highlighting
	newText, newCPos = Misspelled:RemoveHighlighting(text, cPos)

	--If this is a command, don't spellcheck or highlight
	if text:sub(1,1) == "/" then
		if newText ~= text then
			ChatFrameEditBox:SetText(newText)
		end
	end

	--Find misspelled words & populate the WordLocations info.
	Misspelled:CheckLine(newText)

	--Use the WordLocation info to march backwords through the input text,
	--highlighting misspellings
	--tprint(WordLocations)
	for x = #WordLocations, 1, -1 do
		local w = WordLocations[x]
		if WordCache[w.Word].Correct == false then
			--Insert highlighting
			newText = newText:sub(1, w.StartPos -1) .. SPELLED_WRONG_HIGHLIGHT .. newText:sub(w.StartPos, w.EndPos) .. FONT_COLOR_CODE_CLOSE .. newText:sub(w.EndPos+1)

			--Adjust cursor position if needed
		end
	end

	--Adjust the word's WordLocation StartPos and EndPos, wherever we added highlighting.
	--The right click handler uses this position info. to detect the misspelled word that was right clicked.
	--March forward this time
	local n = 0 --NewCharsAddedCounter
	for x = 1, #WordLocations do
		local w = WordLocations[x]
		if WordCache[w.Word].Correct == false then
			w.StartPos = w.StartPos + n + #SPELLED_WRONG_HIGHLIGHT
			n = n + #SPELLED_WRONG_HIGHLIGHT
			w.EndPos = w.EndPos + n
			n = n + #FONT_COLOR_CODE_CLOSE
		end
	end

	if newText ~= text then
		ChatFrameEditBox:SetText(newText)
	end
end



--Finds all words in the input text string, ignoring any Wow hyperlinks or textures
--Next populate the table, WordLocations, storing the following info. on each word:
--	WordLocations[x] = {["word"] = word, ["StartPos"] = pos1, ["EndPos"] = pos2}
--		x == Ordinal position of the word in the text string
--
--Next check each word to see we need to cache spell check info for the word.
--Words with numbers or words in all upper case are ignored.
--Store in the WordCache table whether the word is spell correctly, along with the
--any misspelled replacement suggestions.
--
--Trim the WordCache table if it's grown very large
function Misspelled:CheckLine(text)
	--Reset the info on where each word is located
	WordLocations = {}

	if text == nil then return end
	if #text == 0 then return end

	--Check if we should clear the WordCache table to save memory, if it's gotten very large
	if WordCacheCount > WordCacheCountMax then
		WordCache = {}
		WordCacheCount = 0
	end


	--Find if there are any links on this line, and replace them with the char: #, so they don't match as words
	--in the next phase of parsing.
	-- |c.*|r
	-- |c|H.+|h.+|h|r
	-- |H.*|h
	-- |T.*|t
	-- |n newline character
	local newText = text
	newText = string.gsub(newText, "(|c%x-|H.-|h.-|h|r)", function(x) return string.rep("#", #x) end)
	newText = string.gsub(newText, "(|H.*|h)", function(x) return string.rep("#", #x) end)
	newText = string.gsub(newText, "(|T.*|t)", function(x) return string.rep("#", #x) end)
	newText = string.gsub(newText, "(|n)", function(x) return string.rep("#", #x) end)


	--March through the text, finding the words, record there start & end positions, and spell check status
	local patt = "[A-Za-z0-9_'À-ÿ]+"
	local pos1
	local pos2
	pos1, pos2 = string.find(newText, patt)
	pos = pos1

	local x = 1
	while pos1 ~= nil do
		pos = pos1
		local word = newText:sub(pos1, pos2)

		--ignore all uppercase words
		if word ~= string.upper(word) then
			--ignore words with numbers in them
			if string.match(word, "[%d]") == nil then
				WordLocations[x] = {["Word"] = word, ["StartPos"] = pos1, ["EndPos"] = pos2}

				if WordCache[word] == nil then
					local correct = false

					--Ignore words in all upper case
					if word == string.upper(word) then
						correct = true
					end

					--See if the dictionary contains the word
					if correct == false then
						correct = WordDict:Contains(word)
					end

					--Try the lower case version of the word
					if correct == false then
						correct = WordDict:Contains(string.lower(word))
					end

					--Cache the results
					WordCache[word] = {["Correct"] = correct, ["Suggestions"] = {}}
					WordCacheCount = WordCacheCount + 1
					if correct == false then
						local suggestions = {}
						suggestions = WordDict:Suggest(word)
						if #suggestions > 0 then
							WordCache[word].Suggestions = suggestions
						end
					end
				end
			end
		end
		x = x + 1
		pos1, pos2 = string.find(newText, patt, pos2+1)
	end
end



--Return a string where the highlighting has been removed from any misspelled words,
--with the goal to not distroy any Wow itemLinks or textures. (http://www.wowwiki.com/ItemLink)
--
--It's possible that a recent edit has started to distroy the color tags, either at the
--begginning or end of a highlighted misspelled word.
--Attempt to detect this and remove any dangling tags.
function Misspelled:RemoveHighlighting(text, ...)
	local newText = text

	--Track the number of characters removed from the left side of the current cursor position.
	local cPos
	if ... ~= nil then
		cPos = ...
	else
		cPos = 0
	end

	local itemLinks = {}
	local itemLinkNum = 1

	local patt, pos1, pos2
	local tempText

	--Try and match Item links first.
	patt = "|c%x+|H.-|h.-|h|r"
	pos1, pos2 = string.find(newText, patt)

	while pos1 ~= nil do
		--Store the itemlink and it's relative position so it can be replaced latter
		local itemLink = newText:sub(pos1, pos2)
		itemLinks[itemLinkNum] = itemLink

		--Replace this itemlink with a temporary placeholder code
		newText = string.gsub(newText, patt, "{<<".. tostring(itemLinkNum) .. ">>}", 1)

		itemLinkNum = itemLinkNum + 1
		pos1, pos2 = string.find(newText, patt)
	end

	--Try to match, non colored Item Links
	patt = "|H.-|h"
	pos1, pos2 = string.find(newText, patt)

	while pos1 ~= nil do
		--Store the itemlink and it's relative position so it can be replaced latter
		local itemLink = newText:sub(pos1, pos2)
		itemLinks[itemLinkNum] = itemLink

		--Replace this itemlink with a temporary placeholder code
		newText = string.gsub(newText, patt, "{<<".. tostring(itemLinkNum) .. ">>}", 1)

		itemLinkNum = itemLinkNum + 1
		pos1, pos2 = string.find(newText, patt)
	end

	--Try to match and textures links.  (i.e. Raidtargets and there used when chatting with a GM)
	patt = "|T.-|t"
	pos1, pos2 = string.find(newText, patt)

	while pos1 ~= nil do
		--Store the itemlink and it's relative position so it can be replaced latter
		local itemLink = newText:sub(pos1, pos2)
		itemLinks[itemLinkNum] = itemLink

		--Replace this texture link with a temporary placeholder code
		newText = string.gsub(newText, patt, "{<<".. tostring(itemLinkNum) .. ">>}", 1)

		itemLinkNum = itemLinkNum + 1
		pos1, pos2 = string.find(newText, patt)
	end


	--Remove the highlighting from any misspelled words where the
	--beginning and ending color tags correctly exist.
	patt = SPELLED_WRONG_HIGHLIGHT .. "(.-)|r"
	pos1, pos2 = string.find(newText, patt)
	while pos1 ~= nil do
		tempText = string.gsub(newText, patt, function(x) return x end, 1)

		if #newText - #tempText ~= 0 then
			if cPos >= pos1 then
				cPos = cPos - (#newText - #tempText)
			end
		end
		newText = tempText

		pos1, pos2 = string.find(newText, patt)
	end


	--Remove any remaining orphaned beginning color tags.
	patt = "|c%x%x%x%x%x%x%x%x"
	pos1, pos2 = string.find(newText, patt)
	while pos1 ~= nil do
		tempText = string.gsub(newText, patt, "", 1)

		if #newText - #tempText ~= 0 then
			if cPos >= pos1 then
				cPos = cPos - (#newText - #tempText)
			end
		end
		newText = tempText

		pos1, pos2 = string.find(newText, patt)
	end

	--Remove any remaining orphaned ending color tags.
	patt = "|r"
	pos1, pos2 = string.find(newText, patt)
	while pos1 ~= nil do
		tempText = string.gsub(newText, patt, "", 1)

		if #newText - #tempText ~= 0 then
			if cPos >= pos1 then
				cPos = cPos - (#newText - #tempText)
			end
		end
		newText = tempText

		pos1, pos2 = string.find(newText, patt)
	end

	--print(newText)

	--Replace back the extracted item links
	if #itemLinks > 0 then
		for i,val in ipairs(itemLinks) do
			newText = string.gsub(newText, "{<<" .. tostring(i) .. ">>}", val)
		end
	end


	--Reset the info about where each word is on the line.
	WordLocations = {}

	if cPos < 0 then
		cPos = 0
	end
	return newText, cPos
end



-------------------------------------------------------------------------
--
-- Routines for the right click misspelled suggestions popup.
--
-------------------------------------------------------------------------
function Misspelled_OnMouseUp(self, button)
	if button == "RightButton" then
		--check if we are positioned on a misspelled word
		local pos = ChatFrameEditBox:GetCursorPosition()

		for i, w in ipairs(WordLocations) do
			if	pos >= w.StartPos and pos <= w.EndPos then
				if WordCache[w.Word].Correct == false then
					RightClickedWord = w.Word
					RightClickedWordStartPos = w.StartPos
					RightClickedWordEndPos = w.EndPos

					ToggleDropDownMenu(1, nil, MisspelledSuggestions_DropDown, "cursor")
				else
					RightClickedWord = nil
				end
				break
			end
		end

		--print("cursor position:" , ChatFrameEditBox:GetCursorPosition())
	end
end

function MisspelledSuggestions_InitializeDropDown(level)
	if RightClickedWord == nil then return end
	if #RightClickedWord == 0 then return end

	info = UIDropDownMenu_CreateInfo()
	info.text = "Suggestions for: " .. RightClickedWord
	info.isTitle = 1
	info.notCheckable = true
	UIDropDownMenu_AddButton(info)

	for i, s in ipairs(WordCache[RightClickedWord].Suggestions) do
		info = UIDropDownMenu_CreateInfo()
		info.owner = this:GetParent()
		info.text = s.Word

		--If this suggestion is either a guild member or firend append a note.
		if Misspelled:IsGuildMember(s.Word) == true then
			info.text = info.text .. " (Guild)"
		else
			if Misspelled:IsFriend(s.Word) == true then
				info.text = info.text .. " (Friend)"
			end
		end

		info.isTitle = nil
		info.notCheckable = true
		info.value = s.Word
		info.func = function() SuggestionsFrame_Click(s.Word) end
		--Add the above info to the optios menu as a button
		UIDropDownMenu_AddButton(info)
	end

	--Add a seperator
	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = ""
	info.isTitle = nil
	info.value = ""
	info.notClickable = 1
	info.notCheckable = true
	UIDropDownMenu_AddButton(info)

	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = "Ignore All"
	info.isTitle = nil
	info.value = RightClickedWord
	info.func = function() SuggestionsFrame_Click("###IgnoreAll") end
	info.notClickable = nil
	info.notCheckable = true
	UIDropDownMenu_AddButton(info)

	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = "Add to Dictionary"
	info.isTitle = nil
	info.value = RightClickedWord
	info.func = function() SuggestionsFrame_Click("###AddToDictionary") end
	info.notClickable = nil
	info.notCheckable = true
	UIDropDownMenu_AddButton(info)

	info = UIDropDownMenu_CreateInfo()
	info.owner = this:GetParent()
	info.text = "Cancel"
	info.isTitle = nil
	info.value = nil
	info.notClickable = nil
	info.notCheckable = true
	UIDropDownMenu_AddButton(info)
end

function MisspelledSuggestions_DropDownOnLoad(self)
	UIDropDownMenu_Initialize(self, MisspelledSuggestions_InitializeDropDown, "MENU")
end

function SuggestionsFrame_Click(value)
	--value should equal the suggestion clicked
	--print("Suggestion Clicked: ", value)

	local newText = ChatFrameEditBox:GetText()
	local newCurcorPos = nil

	if value == "###IgnoreAll" then
		--Add this word to the WordCache so it will be ignored as misspelled until you reload
		--print("Ignore:", RightClickedWord)
		WordCache[RightClickedWord].Correct = true
		WordCache[RightClickedWord].Suggestions = {}

		--Remove the misspelled highlighting
		newText = newText:sub(1, RightClickedWordStartPos - 1 - #SPELLED_WRONG_HIGHLIGHT) .. RightClickedWord .. newText:sub(RightClickedWordEndPos + #FONT_COLOR_CODE_CLOSE + 1)
		newCursorPos = RightClickedWordStartPos + #RightClickedWord - #SPELLED_WRONG_HIGHLIGHT - #FONT_COLOR_CODE_CLOSE + 1

	elseif value == "###AddToDictionary" then
		-- Add this word to the user dictionary
		--print("Add to UserDict", RightClickedWord)
		Misspelled:AddToUserDict(RightClickedWord)
		WordCache[RightClickedWord].Correct = true
		WordCache[RightClickedWord].Suggestions = {}

		--Remove the misspelled highlighting
		newText = newText:sub(1, RightClickedWordStartPos - 1 - #SPELLED_WRONG_HIGHLIGHT) .. RightClickedWord .. newText:sub(RightClickedWordEndPos + #FONT_COLOR_CODE_CLOSE + 1)
		newCursorPos = RightClickedWordStartPos + #RightClickedWord - #SPELLED_WRONG_HIGHLIGHT - #FONT_COLOR_CODE_CLOSE + 1

	else
		--replace this word with the selected suggestion.
		--Remove the misspelled highlighting in the process.
		newText = newText:sub(1, RightClickedWordStartPos - 1 - #SPELLED_WRONG_HIGHLIGHT) .. value .. newText:sub(RightClickedWordEndPos + #FONT_COLOR_CODE_CLOSE + 1)

		--save the cursor position just after the corrected word, so we can set it later
		newCursorPos = RightClickedWordStartPos + #value - #SPELLED_WRONG_HIGHLIGHT - #FONT_COLOR_CODE_CLOSE + 1

		--if the replacement word is a different length, then the WordLocations will be messed up.
	end

	ChatFrameEditBox:SetText(newText)

	--printable = gsub(newText, "\124", "\124\124")
    --print("New ChatText:", printable)

	RightClickedWord = nil

	--Recheck the line
	--Misspelled:SpellCheckChat()

	--If we replaced a word, with a suggestion, move the cursor to the end of the new word.
	if newCursorPos ~= nil then
		ChatFrameEditBox:SetCursorPosition(newCursorPos)
	end
end
-----------------------------------------------------
-- End: Right click Suggestions popup
-----------------------------------------------------



-------------------------------------------------------------------------
--
-- Routines for dealing with the user dictionary
--
-------------------------------------------------------------------------

--Load the words saved in the Users dictionary into the baseWords table.
function Misspelled:LoadUserDict()
	if Misspelled_DB == nil then
		Misspelled_DB = {UserDict = {}}
	end

	if Misspelled_DB.UserDict ~= nil then
		for k, v in pairs(Misspelled_DB.UserDict) do
			if WordDict.baseWords[k] == nil then
				WordDict.baseWords[k] = {v[1], v[2]}
			end
		end
	end
	print("Misspelled: User Dictionary Loaded")
end


--Add a new word to the users dictionary, and the currently loaded baseWords table.
--Store both the word and it's phonetic code.
function Misspelled:AddToUserDict(word)
	if word == nil then return end
	if #word == 0 then return end

	if Misspelled_DB.UserDict == nil then
		Misspelled_DB.UserDict = {}
	end
	--Add the new word to the UserDict saved variable
	local pcode = WordDict:PhoneticCode(word)
	Misspelled_DB.UserDict[word] = {"", pcode}

	--And add it to the currently loaded dictionary
	WordDict.baseWords[word] = {"", pcode}

	--Fixup the WordCache
	WordCache[word].Correct = true
	WordCache[word].Suggestions = {}
end
-------------------------------------------------------------------------
-- End: Routines for dealing with the user dictionary
-------------------------------------------------------------------------



-------------------------------------------------------------------------
--
-- Load Guild and Friend Roster Routine
--
-------------------------------------------------------------------------

--Load the player's guild members and friends, as valid words into the
--loaded dictionary.
function Misspelled:LoadGuildAndFriendRoster()

	local numFriends, name
	local pcode

	--print("Misspelled: Guild Members Loading...")

	--First check your friends list
	numFriends = GetNumFriends()
	if numFriends > 0 then
		for i = 1, numFriends do
			name = GetFriendInfo(i)
			if WordDict:Contains(name) == false then
				--Look up the phonetic code for this friend name
				pcode = WordDict:PhoneticCode(name)
				--Add the friend name to the loaded dictionary
				WordDict.baseWords[name] = {"", pcode}
			end
		end
	end

	-- Guild members are valid words.
	if GetNumGuildMembers(true) ~= 0 then
		numFriends = GetNumGuildMembers(true)	-- true to include offline members
		if ( numFriends > 0 ) then
			for i=1, numFriends do
				name = GetGuildRosterInfo(i);
				if name ~= nil then
					if #name ~= 0 then
						if WordDict:Contains(name) == false then
							--Look up the phonetic code for this guild member name
							pcode = WordDict:PhoneticCode(name)
							--Add the guild member to the loaded dictionary
							WordDict.baseWords[name] = {"", pcode}
						end
					end
				end
			end
		end
		--print("Misspelled: Guild Members Loaded")
		this:UnregisterEvent("GUILD_ROSTER_UPDATE")
	end

	--this:UnregisterEvent("GUILD_ROSTER_UPDATE")
end

-------------------------------------------------------------------------
-- End: Load Guild and Friend Roster Routine
-------------------------------------------------------------------------



-------------------------------------------------------------------------
--
-- Utility Routines
--
-------------------------------------------------------------------------

--Split a string, at patt delimiter, into a table
function Misspelled:split(str, patt)
	vals = {}; valindex = 0; word = ""
	-- need to add a trailing separator to catch the last value.
	str = str .. patt
	for i = 1, string.len(str) do

		cha = string.sub(str, i, i)
		if cha ~= patt then
			word = word .. cha
		else
			if word ~= nil then
				vals[valindex] = word
				valindex = valindex + 1
				word = ""
			else
				-- in case we get a line with no data.
				break
			end
		end

	end
	return vals
end

function Misspelled:tprint (t, indent, done)
  -- in case we run it standalone
  local Note = Note or print
--  local Tell = Tell or io.write

  -- show strings differently to distinguish them from numbers
  local function show (val)
    if type (val) == "string" then
      return '"' .. val .. '"'
    else
      return tostring (val)
    end -- if
  end -- show
  -- entry point here
  done = done or {}
  indent = indent or 0
  for key, value in pairs (t) do
    print(string.rep (" ", indent)) -- indent it
    if type (value) == "table" and not done [value] then
      done [value] = true
      Note (show (key), ":");
      tprint (value, indent + 2, done)
    else
      print(show (key), "=")
      print (show (value))
    end
  end
end


--Check your friends list to see if the string: name is on the friends list.
function Misspelled:IsFriend(name)
	local numFriends

	numFriends = GetNumFriends()
	if numFriends > 0 then
		for i = 1, numFriends do
			if name == GetFriendInfo(i) then
				return true
			end
		end
	end
	return false
end

--check the guild roster to see if the string: name, is a guild member.
function Misspelled:IsGuildMember(name)
	local numGuildMembers

	if IsInGuild() then
		numGuildMembers = GetNumGuildMembers(true);	-- true to include offline members
		if ( numGuildMembers > 0 ) then
			for i=1, numGuildMembers do
				if name == GetGuildRosterInfo(i) then
					return true
				end
			end
		end
	end

	return false
end
-------------------------------------------------------------------------
-- End: Utility Routines
-------------------------------------------------------------------------

--[[
--testing
require("bit")
require("WordDict")
--~ require("Dict\\Dic_en_US")
--~ Misspelled_DB = {}
--~ WordDict:Init()
--~ ChatFrameEditBox = {}
--~ function ChatFrameEditBox:GetText()
--~ 	return "Leatherworking"
--~ end
s = "Test |cffff91c8A|cffxxxxxxBSecondWord|r|r"
s2 = string.gsub(s, SPELLED_WRONG_HIGHLIGHT .. "(.-)|r", function(x) return x end)
print(s2)
---]]
